use std::collections::HashMap;
use math::round::ceil;

// common data structures

#[derive(Debug)]
pub struct Task {
    pub id: String,
    pub prio: u8,
    pub deadline: u32,
    pub inter_arrival: u32,
    pub trace: Trace,
}

//#[derive(Debug, Clone)]
#[derive(Debug)]
pub struct Trace {
    pub id: String,
    pub start: u32,
    pub end: u32,
    pub inner: Vec<Trace>,
}

// uselful types

// Our task set
pub type Tasks = Vec<Task>;

// A map from Task/Resource identifiers to priority
pub type IdPrio = HashMap<String, u8>;

// A map from Task identifiers to a set of Resource identifiers
// id_task, {id_res, longest}
pub type TaskResources = HashMap<String, Vec<HashMap<String, u32>>>;

pub type Data = HashMap<String, u32>;
pub type L = HashMap<String, f32>;

// Derives the above maps from a set of tasks
pub fn pre_analysis(tasks: &Tasks) -> (IdPrio, IdPrio, TaskResources, Data, Data) {
    let mut tp = HashMap::new();
    let mut rp = HashMap::new();
    let mut tr: TaskResources = HashMap::new();
    for t in tasks {
        update_prio(t.prio, &t.trace, &mut rp);
		tp.insert(t.id.clone(), t.prio);
        for i in &t.trace.inner {
            update_tr(t.id.clone(), i, &mut tr);
        }
    }
    (tp, rp, tr, inter_arival(&tasks), dead_line(&tasks))
}

// helper functions
fn update_prio(prio: u8, trace: &Trace, hm: &mut IdPrio) {
    if let Some(old_prio) = hm.get(&trace.id) {
        if prio > *old_prio {
            hm.insert(trace.id.clone(), prio);
        }
    } else {
        hm.insert(trace.id.clone(), prio);
    }
    for cs in &trace.inner {
        update_prio(prio, cs, hm);
    }
}

// (s = task id, trace = on res, trmap = tr)
fn update_tr(s: String, trace: &Trace, trmap: &mut TaskResources) {
    if let Some(seen) = trmap.get_mut(&s) {
		let mut hm = HashMap::new();
		let time = trace.end - trace.start;
		let mut found = false;
		for t in seen {
			if let Some(exist) = t.get_mut(&trace.id) {
				if *exist < time {
					*exist = time;
				}
				found = true;
			}
		}
		if !found {
			hm.insert(trace.id.clone(), time);
			trmap.get_mut(&s).unwrap().push(hm);
		}
    } else {
		let mut hm = HashMap::new();
		hm.insert(trace.id.clone(), trace.end - trace.start);
        trmap.insert(s.clone(), vec![hm]);
    }
    for trace in &trace.inner {
        update_tr(s.clone(), trace, trmap);
    }
}

pub fn c_times(tasks: &Tasks) -> Data {
	let mut c = HashMap::new();
	for t in tasks {
		c.insert(t.id.clone(), t.trace.end - t.trace.start);
	}
	c
}

pub fn load(tasks: &Tasks, c: &Data) -> (L, f32){
	let mut l = HashMap::new();
	let mut tot = 0.0;
	for t in tasks {
		let usages = 
			*c.get(&t.id).unwrap() as f32 / t.inter_arrival as f32;
		l.insert(t.id.clone(), usages);
		tot += usages;
	}
	(l, tot)
}

pub fn block(tp:  &IdPrio, rp: &IdPrio, tr: &TaskResources) -> Data {
	let mut b = HashMap::new();
	for (task_id, task_prio) in tp {
		let mut block_time = 0;
		let resourses = tr.get(task_id);		 
		match resourses {
			Some(res) => {
				for vectorn in res {
					for (id, _) in vectorn {
						// iter this, all res
						
						for (tr_id, tr_resorses) in tr {
							if rp.get(tr_id).unwrap() < task_prio {
						
								for bl in tr_resorses {
									if let Some(time) = bl.get(id) {
										block_time += time;
									}
								}
							}
						}
					}
				}
				b.insert(task_id.clone(), block_time);
			},
			None => {
				b.insert(task_id.clone(), 0);
			},
		}

	}
	b
}

pub fn interferens(tp: &IdPrio, c: &Data, a: &Data, dl: &Data, advance: bool) -> Data {
	let mut i = HashMap::new();
	for (id_curent, prio_curent) in tp {
		let mut sum = 0;
		let mut block = *dl.get(&id_curent.clone()).unwrap();
		if advance {
			 block = *c.get(&id_curent.clone()).unwrap();
		} 
		let mut temp = 0;
		while sum < block {
		temp = block;
		let mut part = 0;
			for (id_ref, prio_ref) in tp {
	
				if *id_curent == *id_ref {
					continue;
				} else {
					if *prio_curent == *prio_ref {
						part += *c.get(id_ref).unwrap();
					} else if prio_curent < prio_ref {
						part += block_eq(block, *a.get(id_ref).unwrap(), *c.get(id_ref).unwrap());
					}
				}
			}
			block = part;
			sum = part;
			if !advance {break;}
		}
		i.insert(id_curent.clone(), sum);
	}
	i
}

fn block_eq(block: u32, a: u32, c: u32) -> u32 {
	return ceil(block as f64 / a as f64, 0) as u32 * c;	
}

fn inter_arival(tasks: &Tasks) -> Data {
	let mut a = HashMap::new();
		for t in tasks {
			a.insert(t.id.clone(), t.inter_arrival.clone());
		}
	a
}

pub fn response(c: &Data, b: &Data, i: &Data) -> Data {
	let mut r = HashMap::new();
	for (id, tim) in c {
		r.insert(id.clone(), tim + b.get(id).unwrap() + i.get(id).unwrap());
	}
	r
}

fn dead_line(tasks: &Tasks) -> Data {
	let mut dl = HashMap::new();
	for t in tasks {
		dl.insert(t.id.clone(), t.deadline);
	}
	dl
}
